/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import java.awt.Color;
import java.util.Enumeration;
import java.util.Vector;


public class ColourGroup  {

	private Vector spatialGroups;
	private GroupCollection collection;
	private StackMeasureFile file;
		
	public ColourGroup (SpatialGroup s, GroupCollection collection) {
		spatialGroups = new Vector();
		spatialGroups.add(s);
		this.collection = collection;
		file = s.getFile();
	}
	
	public StackMeasureFile getFile () {
		return file;
	}
	
	public boolean addSpatialGroup (SpatialGroup s) {
		
		// We're going to be quite crude about this
		// to start with.  We're just going to calculate
		// the distance from the new pixel to every
		// pixel we already have and keep it if it matches
		
		Enumeration e = spatialGroups.elements();
		while (e.hasMoreElements()) {
			if (isCloseEnough(s,(SpatialGroup)e.nextElement())) {
				spatialGroups.add(s);
				return true;
			}
		}
		
		return false;
	}
	
	private boolean isCloseEnough (SpatialGroup s1, SpatialGroup s2){
		// We only merge groups from the same file
		if (s1.getFile() != s2.getFile()) return false;
		
		// Check the hue if this is an RGB file
		if (s1.getFile().isRGB()) {
			int hueDiff = s1.getAverageHue()-s2.getAverageHue();
			while (hueDiff < -180) hueDiff += 360;
			while (hueDiff > 180) hueDiff -= 360;
			if (hueDiff < 0) hueDiff = 0-hueDiff;

			//		System.out.println("Compared hue "+s1.getAverageHue()+" to "+s2.getAverageHue()+" with diff "+hueDiff);

			if (hueDiff > collection.getPrefs().hueFlexibility()) return false;
		}
		
		return true;
	}
	
	public int getSpatialGroupCount () {
		return spatialGroups.size();
	}
	
	public Color getAverageColor () {
		int r=0;
		int g=0;
		int b=0;
		Enumeration e = spatialGroups.elements();
		int c;
		for (c=0;e.hasMoreElements();c++) {
			Color col = ((SpatialGroup)e.nextElement()).getBrightestColor();
			r+= col.getRed();
			g+= col.getGreen();
			b+= col.getBlue();
		}
		
		return new Color(r/c,g/c,b/c);
		
	}
	
	public SpatialGroup [] getSpatialGroups () {
		return (SpatialGroup [])spatialGroups.toArray(new SpatialGroup[0]);
	}
	
	public DistanceMeasure getClosestDistance (ColourGroup g2) {
		DistanceMeasure closest = null;
		
		final SpatialGroup [] theseGroups = getSpatialGroups();
		final SpatialGroup [] thoseGroups = g2.getSpatialGroups();
		
		for (int i1=0;i1<theseGroups.length;i1++) {
			for (int i2=0;i2<thoseGroups.length;i2++) {
				DistanceMeasure dist = theseGroups[i1].getClosestDistance(thoseGroups[i2]);
				// No point continuing if we're already at 0!
				if (dist.distance() == 0) {
					return dist;
				}
				if (closest == null || dist.distance() < closest.distance()) {
					closest = dist;
				}
			}
		}
		
		return closest;
		
	}
	
	public DistanceMeasure getCentralDistance (ColourGroup g2) {
		DistanceMeasure closest = null;
		
		final SpatialGroup [] theseGroups = getSpatialGroups();
		final SpatialGroup [] thoseGroups = g2.getSpatialGroups();
		
		for (int i1=0;i1<theseGroups.length;i1++) {
			for (int i2=0;i2<thoseGroups.length;i2++) {
				DistanceMeasure dist = theseGroups[i1].getCentralDistance(thoseGroups[i2]);
				// No point continuing if we're already at 0!
				if (dist.distance() == 0) {
					return dist;
				}
				if (closest == null || dist.distance() < closest.distance()) {
					closest = dist;
				}
			}
		}
		
		return closest;
		
	}
	
}