/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import java.awt.Color;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Vector;

import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;
import java.util.Enumeration;


public class DistanceMeasureSet implements Runnable {

	private final ColourGroup [] colourGroups;
	private final SpatialGroup [] spatialGroups;
	private DistanceMeasure [][] distances;
	private final GroupCollection collection;
	private DistanceTableModel model;

	private boolean stopCalculating = false;
	private boolean isCalculating  = false;
	
	private int groupBy;
	private int measureFrom;
	
	private int totalMeasures;
	private int calculationProgress;
	
	private Vector listeners = new Vector();
	
	// Constants for use in constructor
	public static final int GROUP_BY_COLOUR = 1;
	public static final int GROUP_BY_SPATIAL = 2;
	public static final int MEASURE_BY_PIXEL = 3;
	public static final int MEASURE_FROM_CENTRE = 4;

	
	public DistanceMeasureSet (final GroupCollection collection, int groupBy, int measureFrom) {
		this.groupBy = groupBy;
		this.measureFrom = measureFrom;

		// Check the inputs are OK.
		if (!(groupBy == GROUP_BY_COLOUR || groupBy == GROUP_BY_SPATIAL)) {
			throw new IllegalArgumentException("Group by value not allowed");
		}
		if (!(measureFrom == MEASURE_BY_PIXEL || measureFrom == MEASURE_FROM_CENTRE)) {
			throw new IllegalArgumentException("Measure from value not allowed");
		}		

		this.collection = collection;
		colourGroups = collection.getColourGroups();
		spatialGroups = collection.getSpatialGroups();

		if (groupBy == GROUP_BY_COLOUR) {
			distances = new DistanceMeasure[colourGroups.length][colourGroups.length];
		}
		else {
			distances = new DistanceMeasure[spatialGroups.length][spatialGroups.length];
		}
		
		model = new DistanceTableModel();
		isCalculating = true;
		Thread t = new Thread(this);
		t.start();
	}

	public void addListener (DistanceMeasureListener l) {
		if (l != null && ! listeners.contains(l)) {
			listeners.add(l);
		}
	}

	public void removeListener (DistanceMeasureListener l) {
		if (l != null && listeners.contains(l)) {
			listeners.remove(l);
		}
	}

	
	public void ChangeMeasureOptions (int groupBy, int measureFrom) {
		if (isCalculating) {
			stopCalculating = true;
			while (stopCalculating) {
				// Wait for any thread to die
				try {
					Thread.sleep(20);
				} 
				catch (InterruptedException e) {}
			}
		}
		this.groupBy = groupBy;
		this.measureFrom = measureFrom;
		
		if (groupBy == GROUP_BY_COLOUR) {
			distances = new DistanceMeasure[colourGroups.length][colourGroups.length];
		}
		else {
			distances = new DistanceMeasure[spatialGroups.length][spatialGroups.length];
		}

		model.fireTableStructureChanged();
		
		Thread t = new Thread(this);
		t.start();
	}
	public TableModel getTableModel() {
		return model;
	}
	
	public GroupCollection getCollection () {
		return collection;
	}
	
	public boolean isCalculating () {
		return isCalculating;
	}
	
	public void saveHTMLTable(File file) throws IOException {
		PrintWriter p = new PrintWriter(new FileWriter(file));

		//Start off the html file
		p.println("<html>");
		p.println("<head>");
		p.println("<title>Distances</title>");
		p.println("<style type=\"text/css\">");
		p.println("td {");
		p.println("border-style:solid;");
		p.println("border-width: 1px;");
		p.println("}");
		p.println("</style>");
		p.println("</head>");
		p.println("<body>");
		p.println("<h1>Distances</h1>");
		p.println("<table>");

		// Put out some info on how the values were calculated
		if (groupBy == GROUP_BY_COLOUR) {
			p.println("<h2>Spatial Groups of similar colour were merged</h2>");
		}
		else {
			p.println("<h2>All Spatial Groups are shown separately</h2>");
		}

		if (measureFrom == MEASURE_BY_PIXEL) {
			p.println("<h2>Measures are the closest 2 pixels</h2>");
		}
		else {
			p.println("<h2>Measures are from the centre of mass of each spatial group</h2>");
		}

		//Do the headers
		p.println("<tr>");
		for (int c=0;c<model.getColumnCount();c++) {
			p.println("<th>"+model.getColumnName(c)+"</th>");
		}
		p.println("</tr>");
		
		//Now work our way through the table
		for (int r=0;r<model.getRowCount();r++) {
			p.println("<tr>");
			for (int c=0;c<model.getColumnCount();c++) {
				Object o = model.getValueAt(r, c);
				if (o == null) {
					p.println("<td></td>");
				}
				else if (o instanceof Color) {
					Color col = (Color)o;
					p.println("<td style=\"background:"+getHexString(col)+"\"><pre>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</pre></td>");
				}
				else if (o instanceof DistanceMeasure){
					DistanceMeasure d = (DistanceMeasure)o;
					p.println("<td>"+d.roundedDistance()+"&micro;m</td>");
				}
				else {
					throw new IOException("Don't know how to print "+o.getClass());
				}
			}
			p.println("</tr>");
		}

		p.println("</table></body></html>");

		p.close();

	}
	
	public void saveTextTable(File file) throws IOException {
		PrintWriter p = new PrintWriter(new FileWriter(file));

		// Put out some info on how the values were calculated
		if (groupBy == GROUP_BY_COLOUR) {
			p.println("#merged=1");
		}
		else {
			p.println("#merged=0");
		}

		if (measureFrom == MEASURE_BY_PIXEL) {
			p.println("#measure=pixel");
		}
		else {
			p.println("#measure=centre");
		}

		//Now work our way through the table
		for (int r=0;r<model.getRowCount();r++) {
			for (int c=0;c<model.getColumnCount();c++) {
				if (c != 0){
					p.print("\t");
				}
				
				Object o = model.getValueAt(r, c);
				if (o == null) {
					continue;
				}
				else if (o instanceof Color) {
					int index = c-1;
					if (index == -1) {
						index = r-1;
					}
					if (groupBy == GROUP_BY_COLOUR) {
						p.print(colourGroups[index].getFile().name()+":"+colourGroups[index].getSpatialGroups()[0].getAverageHue());
					}
					else {
						p.print(spatialGroups[index].getFile().name()+":"+spatialGroups[index].getAverageHue());						
					}
				}
				else if (o instanceof DistanceMeasure){
					DistanceMeasure d = (DistanceMeasure)o;
					p.print(d.roundedDistance());
				}
				else {
					throw new IOException("Don't know how to print "+o.getClass());
				}
			}
			p.print("\n");
		}

		p.close();

	}

	private String getHexString (Color c) {
		StringBuffer b = new StringBuffer("#");

		int i = c.getRed();
		String s = Integer.toHexString(i);
		if (s.length() == 1) b.append("0");
		b.append(s);

		i = c.getGreen();
		s = Integer.toHexString(i);
		if (s.length() == 1) b.append("0");
		b.append(s);

		i = c.getBlue();
		s = Integer.toHexString(i);
		if (s.length() == 1) b.append("0");
		b.append(s);

		return b.toString();
	}

	public void run() {
		// Start Calculating Distances

		System.out.println("Starting to calculate distances");

		calculationProgress = 0;
		
		isCalculating = true;
		
		if (groupBy == GROUP_BY_COLOUR) {

			// Work out how many measures we're going to make
			totalMeasures = 0;
			
			for (int i=0;i<colourGroups.length;i++) {
				totalMeasures +=i;
			}
			
			for (int i=0;i<colourGroups.length;i++) {
				try {
					Thread.sleep(10);
				} 
				catch (InterruptedException e1) {
				}
				
				for (int j=i;j<colourGroups.length;j++) {
					calculationProgress++;
					Enumeration e = listeners.elements();
					while (e.hasMoreElements()) {
						((DistanceMeasureListener)e.nextElement()).progressUpdate(calculationProgress, totalMeasures, "Done "+calculationProgress+" out of "+totalMeasures+" measures");
					}
					if (measureFrom == MEASURE_BY_PIXEL) {
						DistanceMeasure d = colourGroups[i].getClosestDistance(colourGroups[j]);
						distances[i][j] = d;
					}
					else {
						DistanceMeasure d = colourGroups[i].getCentralDistance(colourGroups[j]);
						distances[i][j] = d;						
					}
					((AbstractTableModel)model).fireTableDataChanged();
					try {
						if (stopCalculating){
							System.out.println("We've dissapeared.  Better stop calculating");
							stopCalculating = false;
							isCalculating = false;
							return; // We've probably been killed.
						}
						Thread.sleep(20);
					} 
					catch (InterruptedException e2) {}
				}
			}
		}
		else {
			// Work out how many measures we're going to make
			totalMeasures = 0;
			
			for (int i=0;i<spatialGroups.length;i++) {
				totalMeasures +=i;
			}
			for (int i=0;i<spatialGroups.length;i++) {
				
				try {
					Thread.sleep(10);
				} 
				catch (InterruptedException e1) {
				}
				
				for (int j=i;j<spatialGroups.length;j++) {
					calculationProgress++;
					Enumeration e = listeners.elements();
					while (e.hasMoreElements()) {
						((DistanceMeasureListener)e.nextElement()).progressUpdate(calculationProgress, totalMeasures, "Done "+calculationProgress+" out of "+totalMeasures+" measures");
					}
					if (measureFrom == MEASURE_BY_PIXEL) {
						DistanceMeasure d = spatialGroups[i].getClosestDistance(spatialGroups[j]);
						distances[i][j] = d;
					}
					else {
						DistanceMeasure d = spatialGroups[i].getCentralDistance(spatialGroups[j]);
						distances[i][j] = d;						
					}
					((AbstractTableModel)model).fireTableDataChanged();
					try {
						if (stopCalculating){
							System.out.println("We've dissapeared.  Better stop calculating");
							stopCalculating = false;
							isCalculating = false;
							return; // We've probably been killed.
						}
						Thread.sleep(2);
					} 
					catch (InterruptedException e2) {}
				}
			}
		}
		isCalculating = false;
	}

	private class DistanceTableModel extends AbstractTableModel {

		public int getColumnCount() {
			return distances.length+1;
		}

		public int getRowCount() {
			return distances.length+1;
		}

		public Object getValueAt(int row, int col) {
			if (row == 0) {
				if (col == 0) {
					return null;
				}
				else {
					if (groupBy == GROUP_BY_COLOUR) {
						return colourGroups[col-1].getAverageColor();
					}
					else {
						return spatialGroups[col-1].getBrightestColor();						
					}
				}
			}
			else if (col == 0) {
				if (groupBy == GROUP_BY_COLOUR) {
					return colourGroups[row-1].getAverageColor();				
				}
				else {
					return spatialGroups[row-1].getBrightestColor();				
				}
			}
			else {
				return distances[row-1][col-1];
			}
		}

	}

	
}
