/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import ij.ImagePlus;
import ij.io.FileSaver;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.UIManager;
import javax.swing.filechooser.FileFilter;


public class StackMeasureBatchProcessor extends JFrame implements StackMeasureProcessorListener, PreferencesListener, ActionListener, Runnable {

	// This class takes in a set of lsm files and processes them through
	// stack measure.  It offers no chance to review the results, it
	// just chews its way through them.
	
	private JButton runSearchButton;
	private StackMeasurePreferences prefs;
	private BatchPreferences batchPrefs;
	private JList files;
	private DefaultListModel model;
	private boolean prefsOK = true;
	private boolean waiting = true;
	
	
	public StackMeasureBatchProcessor () {
		setTitle("Stack Measure Batch Processor");
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {}
				
		getContentPane().setLayout(new BorderLayout());
		prefs = new StackMeasurePreferences();
		PreferencesPanel optionsPanel = new PreferencesPanel(prefs);
		optionsPanel.addListener(this);
		getContentPane().add(optionsPanel,BorderLayout.EAST);
		
		batchPrefs = new BatchPreferences();
		getContentPane().add(batchPrefs,BorderLayout.WEST);
		
		JPanel filePanel = new JPanel();
		filePanel.setLayout(new BorderLayout());
		filePanel.add(new JLabel("Files",JLabel.CENTER),BorderLayout.NORTH);
		model = new DefaultListModel();
		files = new JList(model);
		filePanel.add(new JScrollPane(files),BorderLayout.CENTER);
		getContentPane().add(filePanel,BorderLayout.CENTER);
		
		JPanel buttonPanel = new JPanel();
		JButton closeButton = new JButton("Close");
		closeButton.setActionCommand("close");
		closeButton.addActionListener(this);
		buttonPanel.add(closeButton);
		
		JButton addFilesButton = new JButton("Add files");
		addFilesButton.setActionCommand("add_files");
		addFilesButton.addActionListener(this);
		buttonPanel.add(addFilesButton);
		
		runSearchButton = new JButton("Start Search");
		runSearchButton.setActionCommand("search");
		runSearchButton.addActionListener(this);
		runSearchButton.setEnabled(false);
		buttonPanel.add(runSearchButton);
		
		getContentPane().add(buttonPanel,BorderLayout.SOUTH);
		
		setSize(600,300);
		setLocationRelativeTo(null);
		setDefaultCloseOperation(JDialog.EXIT_ON_CLOSE);
		
		setVisible(true);

	}
	
	public static void main(String[] args) {
		new StackMeasureBatchProcessor();
	}

	public void exceptionReceived(Exception e) {
		waiting = false;
	}

	public void processingComplete() {
		waiting = false;
	}

	public void progressUpdate(int current, int total, String message) {		
	}

	public void preferencesNotOK() {
		runSearchButton.setEnabled(false);
		prefsOK = false;
	}

	public void preferencesOK() {
		prefsOK = true;
		if (model.getSize() > 0) {
			runSearchButton.setEnabled(true);
		}
	}

	public void actionPerformed(ActionEvent ae) {
		if (ae.getActionCommand().equals("close")) {
			setVisible(false);
			dispose();
		}
		else if (ae.getActionCommand().equals("search")) {
			runSearchButton.setEnabled(false);
			Thread t = new Thread(this);
			t.start();
		}
		else if (ae.getActionCommand().equals("add_files")) {
			JFileChooser chooser = new JFileChooser();
			chooser.setMultiSelectionEnabled(true);
			chooser.setFileFilter(new FileFilter() {
			
				public String getDescription() {
					return "LSM Files";
				}
			
				public boolean accept(File f) {
					if (f.isDirectory() || f.getName().toLowerCase().endsWith(".lsm")) {
						return true;
					}
					else {
						return false;
					}
				}
			
			});
			
			int result = chooser.showOpenDialog(this);
			if (result == JFileChooser.CANCEL_OPTION) return;

			File [] lsmFiles = chooser.getSelectedFiles();

			for (int f=0;f<lsmFiles.length;f++) {
				model.addElement(lsmFiles[f]);
			}
			

			if (model.getSize() > 0 && prefsOK) {
				runSearchButton.setEnabled(true);
			}
		}
		else {
			System.err.println("Unknown action command '"+ae.getActionCommand()+"'");
		}
	}

	public void run() {

		// For each file on the list we're going to make up a GroupCollection.  If
		// needed we'll then make up a distance measure set and then we'll write out
		// the output we need.
		LSM_Reader lsmReader = new LSM_Reader();
		
		BatchProgressDialog prog = new BatchProgressDialog(this);
		
		for (int f=0;f<model.getSize();f++) {
			File file = (File)model.elementAt(f);
			
			prog.progressUpdate(f, model.getSize(), "Reading image "+file.getName());
			
			String baseName = file.getAbsolutePath();
			baseName.replaceAll("\\.lsm$", "");
			
			// First make up a file list based on the layers in this file
			FileList fileList = new FileList();

			ImagePlus [] ips = lsmReader.open(file.getParent(),file.getName(),false,false);
			for (int i=0;i<ips.length;i++) {
				fileList.addFile(new StackMeasureFile(file,ips[i]));
			}
			
			// Now make up a GroupCollection
			GroupCollection collection = new GroupCollection(prefs,fileList);
			StackMeasureProcessor processor = new StackMeasureProcessor(collection);
			processor.addSparkMeasureProcessorListener(prog);
			processor.addSparkMeasureProcessorListener(this);
			waiting = true;
			processor.startProcessing(fileList);
			while (waiting) {
				try {
					Thread.sleep(200);
				} 
				catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			
			// See if they want the spatial output
			if (batchPrefs.spatialOutputWanted()) {
				prog.progressUpdate(0, 1, "Writing spatial output");
				try {
					collection.saveSpatialTextTable(new File (baseName+"_spatial.txt"));
				} 
				catch (IOException e) {
					JOptionPane.showMessageDialog(this, e.getMessage(), "Error writing spatial output", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
				}
			}
			
			// See if they want a composite image
			if (batchPrefs.compositeImageWanted()) {
				prog.progressUpdate(0, 1, "Writing composite image");
				try {
					FileSaver fs = new FileSaver(collection.getFiles().getMergedImage());
					fs.saveAsTiffStack(baseName+"_composite.tiff");
				} 
				catch (Exception e) {
					JOptionPane.showMessageDialog(this, e.getMessage(), "Error writing composite image", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
				}
			}
			
			// See if they want distance output
			if (batchPrefs.distanceOutputWanted()) {
				DistanceMeasureSet dms = new DistanceMeasureSet(collection,batchPrefs.getGroupColoursOption(),batchPrefs.getMeasureFromOption());
				dms.addListener(prog);
				// Since the Distance Measure Set calculates distances in the background we need to wait
				// for it to finish calculating before we try to write out the results.
				
				while (dms.isCalculating()) {
					try {
						Thread.sleep(500);
					} 
					catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				prog.progressUpdate(0, 1, "Writing distance output");
				
				try {
					dms.saveTextTable(new File (baseName+"_distance.txt"));
				} 
				catch (IOException e) {
					JOptionPane.showMessageDialog(this, e.getMessage(), "Error writing distance output", JOptionPane.ERROR_MESSAGE);
					e.printStackTrace();
				}
			}
			
		}
		
		// Once we get here we're finished
		prog.batchProcessingComplete();
		runSearchButton.setEnabled(true);
		JOptionPane.showMessageDialog(this, "Processing completed", "Finished", JOptionPane.INFORMATION_MESSAGE);
		
	}
	
	private class BatchPreferences extends JPanel {
		
		JComboBox groupOutputBox;
		JComboBox distanceOutputBox;
		JComboBox saveImageBox;
		JComboBox groupColoursBox;
		JComboBox measureFromBox;
		
		public BatchPreferences () {
			setLayout(new GridLayout(6,1));
			
			add(new JLabel("Batch Preferences",JLabel.CENTER));
			JPanel groupOutputPanel = new JPanel();
			groupOutputPanel.add(new JLabel("Write spatial group output?"));
			groupOutputBox = new JComboBox(new String [] {"Yes","No"});
			groupOutputPanel.add(groupOutputBox);
			add(groupOutputPanel);

			JPanel distanceOutputPanel = new JPanel();
			distanceOutputPanel.add(new JLabel("Write distance output?"));
			distanceOutputBox = new JComboBox(new String [] {"Yes","No"});
			distanceOutputPanel.add(distanceOutputBox);
			add(distanceOutputPanel);

			JPanel saveImagePanel = new JPanel();
			saveImagePanel.add(new JLabel("Save Composite Image?"));
			saveImageBox = new JComboBox(new String [] {"Yes","No"});
			saveImagePanel.add(saveImageBox);
			add(saveImagePanel);
			
			JPanel groupColoursPanel = new JPanel();
			groupColoursPanel.add(new JLabel("Group colours?"));
			groupColoursBox = new JComboBox(new String [] {"No","Yes"});
			groupColoursPanel.add(groupColoursBox);
			add(groupColoursPanel);
			
			JPanel measureFromPanel = new JPanel();
			measureFromPanel.add(new JLabel("Measure from"));
			measureFromBox = new JComboBox(new String [] {"Centre","Edge"});
			measureFromPanel.add(measureFromBox);
			add(measureFromPanel);
		}
		
		public boolean spatialOutputWanted () {
			if (groupOutputBox.getSelectedItem() == "Yes") {
				return true;
			}
			return false;
		}
		
		public boolean distanceOutputWanted () {
			if (distanceOutputBox.getSelectedItem() == "Yes") {
				return true;
			}
			return false;
		}
		
		public boolean compositeImageWanted () {
			if (saveImageBox.getSelectedItem() == "Yes") {
				return true;
			}
			return false;
		}
		
		public int getGroupColoursOption () {
			if (groupColoursBox.getSelectedItem() == "Yes") {
				return DistanceMeasureSet.GROUP_BY_COLOUR;
			}
			return DistanceMeasureSet.GROUP_BY_SPATIAL;
			
		}

		public int getMeasureFromOption () {
			if (measureFromBox.getSelectedItem() == "Centre") {
				return DistanceMeasureSet.MEASURE_FROM_CENTRE;
			}
			return DistanceMeasureSet.MEASURE_BY_PIXEL;
			
		}
	}

}
